/*
    This file is part of the UNF library / toolset
    Copyright (C) 2000 Steffen Kaiser

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/* $RCSfile: UNF2HC.C $
   $Locker: ska $	$Name:  $	$State: Exp $

	This program is part of the UNF-NLS-kernel implementation.

	UNF2HC - Convert a UNF file into the format of hardcoded NLS pkg

	UNF2HC [file]
	-or-
	UNF2HC country_code codepage_id

	The file is opened and interpreted as an UNF file; if no argument
	is specified, the standard input is read.

	country_code and codepage_id must be decimal numbers.

*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "../version.h"
#include "../unf.h"

#undef fout
#define fout stdout

#ifdef RCS_Version
static char const rcsid[] = 
	"$Id: UNF2HC.C 1.2 2000/07/02 04:51:07 ska Exp ska $";
#endif

static int lineopen = 0, inDB = 0;

/* Dump formatted text */
static void of(char const * const fmt, ...)
{	va_list ap;

	va_start(ap, fmt);
	vfprintf(fout, fmt, ap);
	va_end(ap);
	lineopen = 1;
}
static void os(char const * const str)
{	fputs(str, fout);
	lineopen = 1;
}
static void oc(int ch)
{	putc(ch, fout);
	lineopen = 1;
}

static void onl(void)
{	oc('\n');
	inDB = lineopen = 0;
}
static void oFlush(void)
{	if(lineopen)
		onl();
}
static void oIndent(void)
{	oFlush();
	oc('\t');
}

/* Dump the initial file text */
static void oPrefix(UWORD cntry, UWORD cp)
{
	of("; Hardcoded DOS-NLS information for country = %u, codepage = %u\n"
	 , cntry, cp);
	os("; This is an automatically generated file!\n; Any modifications will be lost!\n\n");
	os("; Prerequisites:\n");
#ifndef NLS_MODIFYABLE_DATA
	os(";; ==> Assuming that data of tables remains constant all the time\n");
#endif
#if defined(DBCS)
	os(";; ==> Reordering tables 1, 2, 4, 5 and 7\n");
#else
	os(";; ==> Reordering tables 1, 2, 4 and 5\n");
#endif
#if defined(FOR_KERNEL)
	os("\n\t%include \"segs.inc\"\nsegment CONST2\n");
#else
	os("\n\t%include \"segs.inc\"\nsegment _DATA\n");
#endif
}
/* Dump the end of file text */
static void oSuffix(void)
{
#if defined(FOR_KERNEL)
	oFlush();
#else
	os("\n\nEND\n");
#endif
}
static void oKeyword(char const * const keyword)
{	char const *p;

	for(p = keyword; *p;)
		oc(toupper(*p++));
	oc(' ');
}

static void oPName(char const * const nam)
{	oc('_');
	os(nam);
}
/* Define a symbol as public */
static void oPubName(char const * const nam)
{	oIndent();
	oKeyword("global");
	oPName(nam);
}

static void oLabelEnd(void)
{	os(":");
	onl();
}
static void oITabName(int id)
{	of("?table%u", id);
}
/* Dump the internal label of a table */
static void oTabLabel(int id)
{
	oFlush();
	oITabName(id);
	oLabelEnd();
}
/* Dump a C label */
static void oCLabel(char const * const nam)
{	if(!nam) return;
	oPubName(nam);
	oFlush();
	oPName(nam);
	oLabelEnd();
}
static void oByte(int b)
{	if(!inDB || inDB >= 8) {
		oIndent();
		oKeyword("db");
	} else
		oc(',');
	of(" 0%02xh", b);
	++inDB;
}
/* Dump a table as serie of DB entries */
static void oTable(UBYTE const *p, int size)
{
	while(size--)
		oByte(*p++);
}
/* Dump a nlsPointer structure to internal table id */
static void oTabPointer(int id)
{	struct nlsPointer poi;

	/* Initialize poi structure */
	poi.subfct = id;
	oTable((UBYTE*)&poi, offsetof(struct nlsPointer, pointer));
	oIndent();
	oKeyword("dw");
	oITabName(id);
	os(", ");
	oKeyword("seg");
	oITabName(id);
}

#define oSubfct(id)	oTabLabel(id);			/* internal label */	\
					oTable(d->data[tab[id]]->data, d->data[tab[id]]->size)

char *hcTabName(int i)
{	switch(i) {
	case 1: return "nlsCountryInfoHardcoded";
	case 2: return "nlsUpcaseHardcoded";
	case 4: return "nlsFUpcaseHardcoded";
	case 5: return "nlsFnameTermHardcoded";
	case 6: return "nlsCollHardcoded";
	case 7: return "nlsDBCSHardcoded";
	}
	return NULL;
}

#define dumpFile(fnam,cntry,cp) dumpFile2(fnam,cntry,cp,0)

static int dumpFile2(char *fnam, WORD cntry, WORD cp, int opt_s)
{	struct UNF_DATA *d;
	int rc, *tab, i, j, maxfct;
	struct nlsPackage pkg;		/* The main entry point of NLS pkgs */

	if((d = unf_load(fnam, cntry, cp, (UNF_Error_Fct *)unf_error)) == NULL) {
		if(!fnam)
			fprintf(stderr, "Failed to load UNF file for %04u:%04u\n"
			 , cntry, cp);
		else if(strcmp(fnam, "-"))
			fprintf(stderr, "Failed to load UNF file from stdin\n");
		else
			fprintf(stderr, "Failed to load UNF file: %s\n", fnam);
		return 33;
	}

	rc = 0;

	if(unf_kernelCheck(d, (UNF_Error_Fct *)unf_error)) {
		rc = 34;
		goto errRet;
	}

		/* The kernelCheck() ensures that the standard tables are
			available: 1, 2, 4, 5, 6, 7, 0x23 */
	tab = unf_tab(d);

	/* Prepare the nlsPackageHardcoded */
	pkg.nxt		= NULL;
	pkg.cntry	= d->country;
	pkg.cp		= d->codepage;
		/* Because this is the hardcoded package, all data is internally
			retreiveble without the need to channel any request through
			MUX */
	pkg.flags	= NLS_FLAG_HARDCODED;
	pkg.yeschar	= d->data[tab[0x23]]->data[0];
	pkg.nochar 	= d->data[tab[0x23]]->data[1];
#if 1
	if (sizeof(((struct nlsPackage *)0)->yeschar) == 2) {
		pkg.yeschar	= *(UWORD *)&(d->data[tab[0x23]]->data[0]);
		pkg.nochar 	= *(UWORD *)&(d->data[tab[0x23]]->data[2]);
	}
#endif
		/* Count the number of tables */
	tab[0x23] = -1;		/* table 0x23 is a pseudo-table */
	pkg.numSubfct = 0;
	i = 256;
	maxfct = 0;		/* highest subfunction number */
	while(i--) if(tab[i] != -1) {
		++pkg.numSubfct;
		if(!maxfct)
			maxfct = i;
	}

	oPrefix(pkg.cntry, pkg.cp);		/* Dump the HC prefix text */

	oCLabel("nlsPackageHardcoded");		/* Dump the main label */
			/* Dump the main body of the main structure */
	oTable((UBYTE*)&pkg, offsetof(struct nlsPackage, nlsPointers));
		/* Now (pkg.numSubfct-1) nlsPointer structures are to be dumped.\
			Because to support NLS_REORDER_POINTERS the tables 2, 4 and
			5 are moved at the beginning and table 1 at the very end of this
			array. */
	oFlush();			/* Separate pointers from other stuff for beauty */
	oTabPointer(2);		/* Dump the default table pointers */
	oTabPointer(4);
	oTabPointer(5);
	if(tab[3] != -1)
		oTabPointer(3);
	for(i = 5; ++i <= maxfct;) if(tab[i] != -1)
		oTabPointer(i);
		/* At the very end, table 1 (Extended Country Info) is added,
			but not as pointer, but plain memory dump */
	oCLabel(hcTabName(1));
	oByte(1);				/* subfct ID preceeds this info block */
#if defined(FOR_KERNEL)
    os("\n\tGLOBAL _nlsCntryInfoHardcoded\n_nlsCntryInfoHardcoded:");
    do {
        UBYTE *d1 = d->data[tab[1]]->data;
        UWORD size1 = d->data[tab[1]]->size;
        
        oTabLabel(1);
        if (size1 > 6) {
            oTable(d1, size1 - 6);
            onl();
            os("extern _CharMapSrvc:wrt DGROUP");
            onl();
            oIndent();
            os("DW  _CharMapSrvc, SEG _CharMapSrvc");
            onl();
            oTable(d1 + (size1 - 6 + 4), 2);
        } else {
            oTable(d1, size1);
        }
    } while (0);
#else
	oSubfct(1);
#endif

	oCLabel("hcTablesStart");

	/* Now dump all the remaining tables and include the HC labels for the
		standard tables */
	for(i=1; ++i <= maxfct;) if(tab[i] != -1) {
#if !defined(NLS_MODIFYABLE_DATA)
# if defined(FOR_KERNEL)
		if (opt_s)
# endif
		/* Check which other tables might use the very same table */
		for(j = i; ++j <= maxfct;) if(tab[j] != -1) {
			/* For simplicity and generality this check does not
				compare both tables by structure (e.g. ignoring
				dummy fields), but compares both tables completely
				byte-per-byte. */
			if(d->data[tab[j]]->size == d->data[tab[i]]->size
			 && memcmp(d->data[tab[j]]->data, d->data[tab[i]]->data
			     , d->data[tab[j]]->size) == 0) {
			     	/* Both tables are completely the same */
			     	/* Place the label of the other table(s) right here,
			     		so both tables will use the same physical
			     		memory */
			     	oCLabel(hcTabName(j));
			     	oTabLabel(j);
			     	tab[j] = -1;	/* this table was dumped already */
			}
		}
#endif
		oCLabel(hcTabName(i));
		oSubfct(i);
	}

	oCLabel("hcTablesEnd");
	oSuffix();			/* Dump the HX suffix text */
	unf_free(tab);

errRet:
	unf_freeData(d);
	return rc;
}


#if 1

static void
usage(const char *progname)
{
    printf("Usage: %s [-s] file.UNF\n", progname);
    printf("       %s [-s] coutry_code codepage_number\n", progname);
    printf("Convert UNF to assembly code for the FreeDOS kernel build 20%02d", REVISION_SEQ);
#if DBCS
    printf(" DBCS");
#endif
    printf("\n\n");
    printf("Option:\n");
    printf("    -s  Shrink duplicate data.\n");
    exit(1);
}

# if !defined(__UNIX__)
static char *
mybasename(char *s)
{
    char *sb, *ss;
    if (!s) return s;
    sb = strrchr(s, '\\');
    sb = sb ? sb + 1 : s;
    ss = strrchr(s, '/');
    ss = ss ? ss + 1 : s;
    s = ss;
    if (s < sb) s = sb;
    return s;
}
# else
#  define mybasename basename
# endif

#define PARAMS 2

int
main(int argc, char **argv)
{

    char *progname;
    char *p;
    char *pa[PARAMS];
    int optH = 0, optS = 0;
    int n = 0;
    int rc;
    
    progname = (argc > 0 && argv[0]) ? mybasename(argv[0]) : "UNF2HC";
    while(argc > 0) {
        --argc;
        p = *++argv;
        if (!p) break;
        if (*p == '-') {
            switch (p[1]) {
            case '?': case 'H': case 'h':
                optH = !0;
                continue;
            case 'S': case 's':
                optS = !0;
                continue;
            case '\0': /* stdin */
                break;
            default:
                fprintf(stderr, "%s: invalid option.\n", progname);
                exit(1);
            }
        }
        if (n < sizeof(pa)/sizeof(pa[0]))
            pa[n++] = p;
    }
    if (optH)
        usage(progname);
    
    switch(n) {
    case 0:     /* no argument */
        rc = dumpFile2("-", UNF_DEFAULT, UNF_DEFAULT, optS);
        break;
    case 1:     /* one argument */
        rc = dumpFile2(pa[0], UNF_DEFAULT, UNF_DEFAULT, optS);
        break;
    case 2:     /* two arguments */
        rc = dumpFile2(NULL, atoi(pa[0]), atoi(pa[1]), optS);
        break;
    default:
        usage(progname);
    }
    
    return rc;
}
#else
main(int argc, char **argv)
{
	switch(argc) {

	case 1:		/* no argument */
		return dumpFile("-", UNF_DEFAULT, UNF_DEFAULT);

	case 2:		/* one argument */
		return dumpFile(argv[1], UNF_DEFAULT, UNF_DEFAULT);

	case 3:		/* two arguments */
		return dumpFile(NULL, atoi(argv[1]), atoi(argv[2]));

	default:
		puts("Useage: UNF2HC file.UNF");
		break;
	}
	return 1;
}
#endif
